/*
   Copyright Universite de Versailles Saint-Quentin en Yvelines 2009
   AUTHORS: Sebastien Briais, Sid Touati

   This file is part of GDD.
   
   GDD is free software: you can redistribute it and/or modify it
   under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation, either version 3 of the
   License, or (at your option) any later version.
   
   GDD is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with GDD.  If not, see
   <http://www.gnu.org/licenses/>.
*/
#include <string.h>
#include <stdio.h>

#include "opcode.h"
#include "util.h"

#include "boxed.h"

/* Include string hash and match functions */
#include "stringutil.h"

/** Create a new opcode .

    @param[in] mnemo = mnemonic
    @param[in] latency = latency 
    @param[in] dflt_dr = default read delay offset

    @return the opcode */
GDD_Opcode *GDD_opcode_create(const char *mnemo, int latency, int dflt_dr) {
  GDD_Opcode *op = (GDD_Opcode *)safe_malloc(sizeof(GDD_Opcode));
  op->mnemo = safe_strdup(mnemo);
  op->latency = latency;
  op->dflt_dr = dflt_dr;
  op->delta_r = SCEDA_hashmap_create((SCEDA_delete_fun)free,
				     (SCEDA_delete_fun)boxed_delete,
				     (SCEDA_match_fun)string_match,
				     (SCEDA_hash_fun)string_hash);
  op->delta_w = SCEDA_hashmap_create((SCEDA_delete_fun)free,
				     (SCEDA_delete_fun)boxed_delete,
				     (SCEDA_match_fun)string_match,
				     (SCEDA_hash_fun)string_hash);
  return op;
}

/** Delete an opcode.

    @param[in] op = opcode */
void GDD_opcode_delete(GDD_Opcode *op) {
  SCEDA_hashmap_delete(op->delta_r);
  SCEDA_hashmap_delete(op->delta_w);
  free(op->mnemo);
  free(op);
}

/** Extend delta_r function (read delay offset) of given opcode.

    @param[in] op = opcode
    @param[in] type = considered type
    @param[in] read = read delay offset for type

    @return 0 in case of success, or -1 in case of failure (e.g. if
    trying to redefine delta_r with another value).
*/
int GDD_opcode_extend_delta_r(GDD_Opcode *op, const char *type, int read) {
  if(SCEDA_hashmap_contains_key(op->delta_r, type)) {
    fprintf(stderr,"Warning: delta_r(%s, %s) multiply defined\n", GDD_opcode_mnemonic(op), type);
    boxed(int) oread = SCEDA_hashmap_get(op->delta_r, type);
    if(boxed_get(oread) == read) {
      return 0;
    } else {
      return -1;
    }
  } else {
    return SCEDA_hashmap_put(op->delta_r,safe_strdup(type), boxed_create(int, read),NULL);
  }
}

/** Extend delta_w function (write delay offset) of given opcode.

    @param[in] op = opcode
    @param[in] type = considered type
    @param[in] write = write delay offset for type

    @return 0 in case of success, or -1 in case of failure (e.g. if
    trying to redefine delta_r with another value).
*/
int GDD_opcode_extend_delta_w(GDD_Opcode *op, const char *type, int write) {
  if(SCEDA_hashmap_contains_key(op->delta_w, type)) {
    fprintf(stderr,"Warning: delta_w(%s, %s) multiply defined\n", GDD_opcode_mnemonic(op), type);
    boxed(int) owrite = SCEDA_hashmap_get(op->delta_w, type);
    if(boxed_get(owrite) == write) {
      return 0;
    } else {
      return -1;
    }
  } else {
    return SCEDA_hashmap_put(op->delta_w,safe_strdup(type), boxed_create(int, write),NULL);
  }
}

/** Get the value of read delay offset 

    @param[in] op = opcode
    @param[in] type = considered type

    @return read delay offset (or default read delay offset) 
    if type == NULL, return default read delay offset
*/
int GDD_opcode_delta_r(GDD_Opcode *op, const char *type) {
  if(type == NULL) {
    return op->dflt_dr;
  }
  if(SCEDA_hashmap_contains_key(op->delta_r, type)) {
    boxed(int) read = SCEDA_hashmap_get(op->delta_r, type);
    return boxed_get(read);
  } else {
    return op->dflt_dr;
  }
}

/** Get the value of write delay offset 

    @param[in] op = opcode
    @param[in] type = considered type

    @return write delay offset (0 by default) */
int GDD_opcode_delta_w(GDD_Opcode *op, const char *type) {
  if(SCEDA_hashmap_contains_key(op->delta_w, type)) {
    boxed(int) write = SCEDA_hashmap_get(op->delta_w, type);
    return boxed_get(write);
  } else {
    return 0;
  }
}

/** Test whether an opcode is a value of given type.

    @param[in] op = opcode
    @param[in] type = type

    @return TRUE if op if a value of type type, FALSE otherwise */
int GDD_opcode_is_value(GDD_Opcode *op, const char *type) {
  return SCEDA_hashmap_contains_key(op->delta_w, type);
}

/** Print the description of an opcode, including its delay offsets.

    @param[in] stream = file stream where to print
    @param[in] op = opcode to print */
void GDD_opcode_fprint(FILE *stream, GDD_Opcode *op) {
  fprintf(stream, "opcode mnemo = %s\n", GDD_opcode_mnemonic(op));
  fprintf(stream, "latency = %d\n", GDD_opcode_latency(op));
  fprintf(stream, "default dr = %d\n", op->dflt_dr);

  {
    SCEDA_HashMapIterator iter;
    SCEDA_hashmap_iterator_init(op->delta_r, &iter);
    while(SCEDA_hashmap_iterator_has_next(&iter)) {
      char *type;
      boxed(int) v = SCEDA_hashmap_iterator_next(&iter, &type);
      fprintf(stream, "delta_r(%s) = %d\n", type, boxed_get(v));
    }
    SCEDA_hashmap_iterator_cleanup(&iter);
  }

  {
    SCEDA_HashMapIterator iter;
    SCEDA_hashmap_iterator_init(op->delta_w, &iter);
    while(SCEDA_hashmap_iterator_has_next(&iter)) {
      char *type;
      boxed(int) v = SCEDA_hashmap_iterator_next(&iter, &type);
      fprintf(stream, "delta_w(%s) = %d\n", type, boxed_get(v));
    }
    SCEDA_hashmap_iterator_cleanup(&iter);
  }
}
