/*
   Copyright Universite de Versailles Saint-Quentin en Yvelines 2009
   AUTHORS: Sebastien Briais, Sid Touati

   This file is part of RS.
   
   RS is free software: you can redistribute it and/or modify it
   under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation, either version 3 of the
   License, or (at your option) any later version.
   
   RS is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with RS.  If not, see
   <http://www.gnu.org/licenses/>.
*/
#include <sys/time.h>

#include <stdio.h>
#include <libxml/parser.h>

#include <stdarg.h>

#include <assert.h>

#include <GDD/arch_xml.h>
#include <GDD/ddg_xml.h>
#include <RS/greedy.h>
#include <RS/exact.h>

static double timeout = 3600;

void usage(const char *app) {
  fprintf(stdout,"%s -a arch.xml -i ddg.xml [-t regtype] [-exact] [-stats] [--timeout #value]\n",app);
}

void warn_error(int status, const char *app, const char *fmt, ...) {
  va_list ap;
  if(status == 0) {
    fprintf(stderr,"warning: ");
  } else {
    fprintf(stderr,"error: ");
  }
  va_start(ap, fmt);
  vfprintf(stderr, fmt, ap);
  va_end(ap);
  fprintf(stderr,"\n");
  usage(app);
  if(status != 0) {
    exit(status);
  }
}

double ellapsed_time(struct timeval *t0, struct timeval *t1) {
  return ((double)(t1->tv_sec - t0->tv_sec)) + ((double)(t1->tv_usec - t0->tv_usec))/((double)1000000);
}

void real_saturation(GDD_DAG *ddg, const char *type) {
  int sat;
  SCEDA_HashMap *k;
  int exact = RS_ddag_compute_saturation(ddg, type, timeout, &sat, &k);

  if(exact) {
    fprintf(stdout, "%s saturation (exact) = %d\n", type, sat);
  } else {
    if(k != NULL) {
      fprintf(stdout, "%s saturation (exact-aborted) = %d\n", type, sat);
    }
  }

  if(k != NULL) {
    SCEDA_hashmap_delete(k);
  }
}

void approx_saturation(GDD_DAG *ddg, const char *type) {
  int n = RS_ddag_estimate_saturation(ddg, type, TRUE); // fast version

  if(n > 0) {
    fprintf(stdout, "%s saturation = %d\n", type, n); 
  }
}

void do_stats(GDD_DAG *ddg, const char *type) {
  int n = GDD_nb_values(GDD_dag_get_ddg(ddg), type);
  if(n > 0) {
    fprintf(stdout,"%s values = %d\n", type, n);
    fprintf(stdout,"%s flow edges = %d\n", type, GDD_nb_flowdeps(GDD_dag_get_ddg(ddg), type));
  }
}

int main(int argc, char *argv[]) {
  GDD_Arch *arch = NULL;
  SCEDA_Graph *dag = NULL;
  char *type = NULL;

  void (*action)(GDD_DAG *ddg, const char *type) = approx_saturation;

  xmlInitParser();

  int i;
  for(i = 0; i < argc; i++) {
    if(strcmp(argv[i],"-i") == 0) {
      i++;
      if(i < argc) {
	if(dag == NULL) {
	  char *filename = argv[i];
	  if(arch == NULL) {
	    fprintf(stderr,"warning: no arch file given\n");
	    fprintf(stderr,"attempting to load default arch file\n");
	    arch = GDD_arch_of_xml_file("default_arch.xml");
	    assert(arch != NULL);
	  }
	  fprintf(stderr,"loading ddg file %s\n",filename);
	  dag = GDD_of_xml_file(TRUE, arch, filename, NULL);
	  assert(dag != NULL);
	} else {
	  warn_error(1, argv[0], "ddg file already given");
	}
      }
    } else if(strcmp(argv[i],"-a") == 0) {
      i++;
      if(i < argc) {
	if(arch == NULL) {
	  char *filename = argv[i];
	  fprintf(stderr,"loading arch file %s\n",filename);
	  arch = GDD_arch_of_xml_file(filename);
	  assert(arch != NULL);
	} else {
	  warn_error(1, argv[0], "arch file already given");
	}
      }
    } else if(strcmp(argv[i],"-t") == 0) {
      i++;
      if(i < argc) {
	type = argv[i];
      } else {
	warn_error(1, argv[0], "missing register type");
      }
    } else if(strcmp(argv[i],"-exact") == 0) {
      action = real_saturation;
    } else if(strcmp(argv[i],"-stats") == 0) {
      action = do_stats;
    } else if(strcmp(argv[i],"--timeout") == 0) {
      i++;
      if(i < argc) {
	timeout = atof(argv[i]);
      } else {
	warn_error(1, argv[0], "missing timeout value");
      }
    }
  }

  xmlCleanupParser();

  if(dag == NULL) {
    warn_error(1, argv[0], "no ddg file given");
  }

  fprintf(stderr,"building GDD_DAG internal structure... ");
  GDD_DAG *ddg = GDD_dag_create(dag); 
  assert(ddg != NULL);
  fprintf(stderr,"done.\n");

  //
  struct timeval t0;
  gettimeofday(&t0, NULL);

  if(type == NULL) {
    GDD_ArchTypesIterator all_types;
    GDD_arch_types_iterator_init(arch, &all_types);
    while(GDD_arch_types_iterator_has_next(&all_types)) {
      type = GDD_arch_types_iterator_next(&all_types);
      action(ddg, type);
    }
    GDD_arch_types_iterator_cleanup(&all_types);
  } else {
    action(ddg, type);
  }

  struct timeval t1;
  gettimeofday(&t1, NULL);

  fprintf(stdout,"ellapsed time = %f\n",ellapsed_time(&t0, &t1));
  //

  GDD_dag_delete(ddg);

  SCEDA_graph_delete(dag);
  GDD_arch_delete(arch);

  return 0;
}
