/*
   Copyright Universite de Versailles Saint-Quentin en Yvelines 2009
   AUTHORS: Sebastien Briais, Sid Touati

   This file is part of siralina.
   
   siralina is free software: you can redistribute it and/or modify it
   under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation, either version 3 of the
   License, or (at your option) any later version.
   
   siralina is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with siralina.  If not, see
   <http://www.gnu.org/licenses/>.
*/
#include <stdio.h>
#include <stdlib.h>
#include <libxml/parser.h>

#include <stdarg.h>

#include <GDD/arch_xml.h>
#include <GDD/ddg_xml.h>

#include <SIRA/loop.h>
#include <SIRA/solver.h>

#include <assert.h>

#include <SIRA/isolver.h>

#define safe_strdup(s) (strdup(s))

#define safe_call(x) (x)

void usage(const char *app) {
  fprintf(stdout,"%s -a arch.xml -i ddg.xml", app);
  fprintf(stdout," [-t regtype #value | {-y regtype}]");
  fprintf(stdout," [--IImin #value] [--IImax #value] [--II #value]");
  fprintf(stdout," [--ual] [--bounded] [--fix | --lao #value | --sira]");
  fprintf(stdout," [--check]");
  fprintf(stdout," [--Lmax #value]");
  fprintf(stdout," [--max-iter #value]");
  fprintf(stdout,"\n");
}

#define WARNING 0
#define ERROR 1

void warn_error(int status, const char *app, const char *fmt, ...) {
  va_list ap;
  if(status == WARNING) {
    fprintf(stderr,"warning: ");
  } else {
    fprintf(stderr,"error: ");
  }
  va_start(ap, fmt);
  vfprintf(stderr, fmt, ap);
  va_end(ap);
  fprintf(stderr,"\n");
  usage(app);
  if(status != WARNING) {
    exit(status);
  }
}

int main(int argc, char *argv[]) {
  GDD_Arch *arch = NULL;
  SCEDA_Graph *g = NULL;

  int IImin_rec;
  int IImin = -1;
  int IImax = -1;
  int L = -1;

  SCEDA_List *types = SCEDA_list_create((SCEDA_delete_fun)free);

  SIRA_Param settings;
  SIRA_param_init(&settings);
  SIRA_param_set_ual_flag(&settings, FALSE);
  SIRA_param_set_bounded_flag(&settings, FALSE);
  SIRA_param_set_check_flag(&settings, FALSE);
  SIRA_param_set_max_iter(&settings, 0);

  SIRA_WeightMap *weight_map = SIRA_weight_map_create();

  int stats_mode = FALSE;

  SIRA_LMethod lmeth = Lfix;
  int lao_pdeg = 1;

  int compute_MII = FALSE;

  xmlInitParser();

  int i;
  for(i = 0; i < argc; i++) {
    if(strcmp(argv[i],"-i") == 0) {
      i++;
      if(i < argc) {
	if(g == NULL) {
	  char *filename = argv[i];
	  if(arch == NULL) {
	    fprintf(stderr,"warning: no arch file given\n");
	    fprintf(stderr,"attempting to load default arch file\n");
	    arch = GDD_arch_of_xml_file("default_arch.xml");
	    assert(arch != NULL);
	  }
	  fprintf(stderr,"loading ddg file %s\n",filename);
	  g = GDD_of_xml_file(FALSE, arch, filename, &IImin_rec);
	  assert(g != NULL);
	} else {
	  warn_error( ERROR, argv[0], "ddg file already given");
	}
      } 
    } else if(strcmp(argv[i],"-a") == 0) {
      i++;
      if(i < argc) {
	if(arch == NULL) {
	  char *filename = argv[i];
	  fprintf(stderr,"loading arch file %s\n",filename);
	  arch = GDD_arch_of_xml_file(filename);
	  assert(arch != NULL);
	} else {
	  warn_error( ERROR, argv[0], "arch file already given");
	}
      }
    } else if(strcmp(argv[i],"-t") == 0) {
      i++;
      if(i < argc) {
	if(SCEDA_list_size(types) != 0) {
	  warn_error( ERROR, argv[0], "a list of types is already defined");
	} else {
	  char *type = argv[i];
	  i++;
	  if(i < argc) {
	    int w = atoi(argv[i]);
	    if(w < 0) {
	      warn_error( ERROR, argv[0], "invalid weight %d", w);
	    } else {
	      if(SIRA_weight_map_set_weight_of_type(weight_map, type, w) != 0) {
		warn_error( ERROR, argv[0], "problem when setting weight of type %s", type);
	      }
	    }
	  } else {
	    warn_error( ERROR, argv[0], "missing weight");
	  }
	}
      } else {
	warn_error( ERROR, argv[0], "missing register type");
      }
    } else if(strcmp(argv[i],"-y") == 0) {
      i++;
      if(i < argc) {
	if(!SIRA_weight_map_is_empty(weight_map)) {
	  warn_error( ERROR, argv[0], "a type -> weight map is already defined");
	} else {
	  char *type= argv[i];
	  safe_call(SCEDA_list_add_tail(types, safe_strdup(type)));
	}
      } else {
	warn_error( ERROR, argv[0], "missing register type");
      }
    } else if(strcmp(argv[i],"--IImin") == 0) {
      i++;
      if(i < argc) {
	IImin = atoi(argv[i]);
	if(IImin < 0) {
	  warn_error( WARNING, argv[0], "illegal IImin value %d", IImin);
	}
      } else {
	warn_error( ERROR, argv[0], "missing IImin value");
      }
    } else if(strcmp(argv[i],"--IImax") == 0) {
      i++;
      if(i < argc) {
	IImax = atoi(argv[i]);
	if(IImax < 0) {
	  warn_error( WARNING, argv[0], "illegal IImax value %d", IImax);
	}
      } else {
	warn_error( ERROR, argv[0], "missing IImax value");
      }
    } else if(strcmp(argv[i],"--II") == 0) {
      i++;
      if(i < argc) {
	IImin = atoi(argv[i]);
	IImax = IImin;
	if(IImin < 0) {
	  warn_error( WARNING, argv[0], "illegal II value %d", IImax);
	}
      } else {
	warn_error( ERROR, argv[0], "missing II value");
      }
    } else if(strcmp(argv[i],"--ual") == 0) {
      SIRA_param_set_ual_flag(&settings, TRUE);
    } else if(strcmp(argv[i],"--bounded") == 0) {
      SIRA_param_set_bounded_flag(&settings, TRUE);
    } else if(strcmp(argv[i],"--fix") == 0) {
      lmeth = Lfix;
    } else if(strcmp(argv[i],"--sira") == 0) {
      lmeth = Lsira;
    } else if(strcmp(argv[i],"--lao") == 0) {
      lmeth = Llao;
      i++;
      if(i < argc) {
	lao_pdeg = atoi(argv[i]);
	if(lao_pdeg <= 0) {
	  warn_error( ERROR, argv[0], "illegal pipelining degree %d", lao_pdeg);
	}
      } else {
	warn_error( ERROR, argv[0], "missing pipelining degree");
      }
    } else if(strcmp(argv[i],"--Lmax") == 0) {
      i++;
      if(i < argc) {
	L = atoi(argv[i]);
	if(L < 0) {
	  warn_error( WARNING, argv[0], "illegal Lmax value %d", L);
	}
      } else {
	warn_error( ERROR, argv[0], "missing Lmax value");
      }
    } else if(strcmp(argv[i],"--MII") == 0) {
      compute_MII = TRUE;
    } else if(strcmp(argv[i],"--check") == 0) {
      SIRA_param_set_check_flag(&settings, TRUE);
    } else if(strcmp(argv[i],"--statistics") == 0) {
      stats_mode = TRUE;
    } else if(strcmp(argv[i],"--max-iter") == 0) {
      i++;
      if(i < argc) {
	int n = atoi(argv[i]);
	if(n < 0) {
	  warn_error( ERROR, argv[0], "illegal maximal number of iterations %d", n);
	}
	SIRA_param_set_max_iter(&settings, n);
      } else {
	warn_error( ERROR, argv[0], "missing maximal number of iterations value");
      }
    }
  }

  xmlCleanupParser();

  if(g == NULL) {
    warn_error( ERROR, argv[0], "no ddg file given");
  }

  if(compute_MII) {
    int MIInum, MIIden;
    SCEDA_List *cycle;
    if(GDD_MII_ratio_cycle(g, &MIInum, &MIIden, &cycle) == 0) {
      fprintf(stdout,"%d/%d\n", MIInum, MIIden);
      SCEDA_list_delete(cycle);
    }
    SIRA_param_cleanup(&settings);
    SCEDA_graph_delete(g);
    GDD_arch_delete(arch);
    return 0;
  }

  SIRA_Loop *loop = SIRA_loop_create(g, IImin_rec);
  SIRA_loop_set_L(loop, L);

  switch(lmeth) {
  case Lfix:
    SIRA_loop_set_fixed_method(loop);
    break;
  case Lsira:
    SIRA_loop_set_sira_method(loop);
    break;
  case Llao:
    SIRA_loop_set_lao_method(loop, lao_pdeg);
    break;
  default:
    SIRA_loop_set_fixed_method(loop);
    break;
  }

  SIRA_RegConstraintsMap *regconstraints = SIRA_regconstraints_map_create();
  {
    GDD_ArchTypesIterator all_types;
    GDD_arch_types_iterator_init(arch, &all_types);
    while(GDD_arch_types_iterator_has_next(&all_types)) {
      char *type = GDD_arch_types_iterator_next(&all_types);
      int n = GDD_arch_get_regtype(arch, type);
      fprintf(stderr,"Max register of type %s = %d\n",type,n);
      SIRA_regconstraints_map_set_constraint_of_type(regconstraints, type, n);
    }
    GDD_arch_types_iterator_cleanup(&all_types);
  }
  

  if(SCEDA_list_size(types) != 0) {
    if(stats_mode) {
      SIRA_isolver_stats(loop, &settings, types, IImin, IImax);
    } else {
      SCEDA_HashMap *solution;
      SCEDA_HashMap *costmap;
      int II = SIRA_isolver_search(loop, regconstraints, &settings, types, IImin, IImax, &solution, &costmap);
      if(II >= 0) {
	fprintf(stdout,"siralina found a (sequential) solution for II = %d\n", II);
	if(SIRA_param_get_ual_flag(&settings)) {
	  fprintf(stdout,"UAL semantics\n");
	} else {
	  fprintf(stdout,"NUAL semantics\n");
	}
	SIRA_isolver_undo(loop, solution);
	SCEDA_hashmap_delete(solution);
	SCEDA_hashmap_delete(costmap);
      } else {
	fprintf(stdout,"siralina did not find any (sequential) solution.\n");
      }
    }
  } else {
    if(SIRA_weight_map_is_empty(weight_map)) {
      GDD_ArchTypesIterator all_types;
      GDD_arch_types_iterator_init(arch, &all_types);
      while(GDD_arch_types_iterator_has_next(&all_types)) {
	char *type = GDD_arch_types_iterator_next(&all_types);
	SIRA_weight_map_set_weight_of_type(weight_map, type, 1);
      }
      GDD_arch_types_iterator_cleanup(&all_types);
    }
    
    if(stats_mode) {
      SIRA_solver_stats(loop, &settings, weight_map, IImin, IImax);
    } else {
      SCEDA_HashMap *solution;
      int II = SIRA_solver_search(loop, regconstraints, &settings, weight_map, IImin, IImax, &solution);
      if(II >= 0) {
	fprintf(stdout,"siralina found a solution for II = %d\n", II);
	if(SIRA_param_get_ual_flag(&settings)) {
	  fprintf(stdout,"UAL semantics\n");
	} else {
	  fprintf(stdout,"NUAL semantics\n");
	}
	SCEDA_hashmap_delete(solution);
      } else {
	fprintf(stdout,"siralina did not find any solution.\n");
      }
    }
  }

  SIRA_param_cleanup(&settings);
  SIRA_weight_map_delete(weight_map);

  SIRA_regconstraints_map_delete(regconstraints);

  SIRA_loop_delete(loop);

  SCEDA_graph_delete(g);
  GDD_arch_delete(arch);

  SCEDA_list_delete(types);

  return 0;
}
