/*
   Copyright Universite de Versailles Saint-Quentin en Yvelines 2009
   AUTHORS: Sebastien Briais, Sid Touati

   This file is part of SIRA.
   
   SIRA is free software: you can redistribute it and/or modify it
   under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation, either version 3 of the
   License, or (at your option) any later version.
   
   SIRA is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with SIRA.  If not, see
   <http://www.gnu.org/licenses/>.
*/
#ifndef __SIRA_ISOLVER_H
#define __SIRA_ISOLVER_H
/** \file isolver.h
    \brief Sequential SiraLina solver */

#include <SCEDA/list.h>
#include <SCEDA/hashmap.h>

#include "loop.h"
#include "param.h"
#include "regconstraints.h"

/** Undo the modifications performed by the sequential siralina solver.

    @param[in] loop = siraloop
    @param[in] seq_solution = sequential siralina solution to undo

    Note that the hashmap is cleared but not deleted. */
void SIRA_isolver_undo(SIRA_Loop *loop, SCEDA_HashMap *seq_solution);

/** Apply sequential siralina heuristic for given II.

    @param[in] loop = siraloop
    @param[in] settings = solver settings (semantics might be modified)
    @param[in] types = list of types to process in specified order
    @param[in] II = II (if negative, fails)
    @param[out] seq_solution = computed solution (map from types to antidep edges list)
    @param[out] costmap = cost of solution (map from types to integer pointers)

    @return 0 if a solution was found, otherwise -1 
    Beware that the loop is modified as a side-effect when a solution is found. */
int SIRA_isolver_single(SIRA_Loop *loop, SIRA_Param *settings, SCEDA_List *types, int II, SCEDA_HashMap **seq_solution, SCEDA_HashMap **costmap);

/** Apply sequential siralina heuristic for every II between IImin and
    IImax and stops when a solution is found (ie satisfies the
    register constraints).

    @param[in] loop = siraloop
    @param[in] constraints = register constraints
    @param[in] settings = solver settings (semantics might be modified)
    @param[in] types = list of types to process in specified order
    @param[in] IImin = IImin (if negative, IImin of the loop will be used)
    @param[in] IImax = IImax (if negative, IImax of the loop will be used) 
    @param[out] seq_solution = computed solution (map from types to antidep edges list)
    @param[out] costmap = cost of solution (map from types to integer pointers)

    @return II >= 0 if a solution was found, otherwise -1 
    Beware that the loop is modified as a side-effect when a solution is found. */
int SIRA_isolver_search(SIRA_Loop *loop, SIRA_RegConstraintsMap *constraints, SIRA_Param *settings, SCEDA_List *types, int IImin, int IImax, SCEDA_HashMap **seq_solution, SCEDA_HashMap **costmap);

/** Apply sequential siralina heuristic for every II between IImin and IImax and
    print the results (used for statistics).

    @param[in] loop = siraloop
    @param[in] settings = solver settings
    @param[in] types = list of types to process in specified order
    @param[in] IImin = IImin (if negative, IImin of the loop will be used)
    @param[in] IImax = IImax (if negative, IImax of the loop will be used) */
void SIRA_isolver_stats(SIRA_Loop *loop, SIRA_Param *settings, SCEDA_List *types, int IImin, int IImax);

#endif

/** \page isolver Sequential SiraLina solver 

    \section isolver_intro Introduction

    As we have explained, \ref solver is able to solve problem
    involving multiple register types by treating all types
    simultaneously.

    However, when more precise information is known about register
    types and constraints, it might be interesting to focus on certain
    types before others. 

    The sequential SiraLina solver thus allows to solve problem
    involving multiple register types by treating each type
    sequentially, in the desired order.

    \section isolver_api Sequential solver API

    As for \ref solver, solver settings are set via a SIRA_Param
    structure. See also \ref param_api.

    There are also three modes of operation for the sequential solver.

    Note that contrary to \ref solver, the sequential solver will
    modify the input DDG to reflect the changes associated to the
    reuse edges. 

    Thus an undo function is provided to undo the applied
    modifications.

    \code
    void SIRA_isolver_stats(SIRA_Loop *loop, SIRA_Param *settings, SCEDA_List *types, int IImin, int IImax);
    \endcode

    This mode is used for statistical analysis. It applies sequential
    siralina solver for every II between IImin and IImax and prints
    the results.

    Note that the input DDG is <b>not</b> modified.

    \code
    int SIRA_isolver_search(SIRA_Loop *loop, SIRA_RegConstraintsMap *constraints, SIRA_Param *settings, SCEDA_List *types, int IImin, int IImax, SCEDA_HashMap **seq_solution, SCEDA_HashMap **costmap);
    \endcode

    In this mode, the solver search for the first II (between IImin
    and IImax) that has a solution which satisfies the register
    constraints.

    Note that the input DDG is <b>modified</b>. Added edges are
    bookkept in the seq_solution map.

    \code
    int SIRA_isolver_single(SIRA_Loop *loop, SIRA_Param *settings, SCEDA_List *types, int II, SCEDA_HashMap **seq_solution, SCEDA_HashMap **costmap);
    \endcode

    In this mode, the siralina solver computes a solution for given II.

    Note that the input DDG is <b>modified</b>. Added edges are
    bookkept in the seq_solution map.

    \code
    void SIRA_isolver_undo(SIRA_Loop *loop, SCEDA_HashMap *seq_solution);
    \endcode

    This function reverts the modifications that were made by the
    sequential siralina solver procedure (see above). It deletes the
    antidep edges listed in the given map and clears the map.

    \section isolver_exa Example

    \include "isolver/main.c"

 */
