/*
   Copyright Universite de Versailles Saint-Quentin en Yvelines 2009
   AUTHORS: Sebastien Briais, Sid Touati

   This file is part of SIRA.
   
   SIRA is free software: you can redistribute it and/or modify it
   under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation, either version 3 of the
   License, or (at your option) any later version.
   
   SIRA is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with SIRA.  If not, see
   <http://www.gnu.org/licenses/>.
*/
#include <SCEDA/graph_dag.h>
#include <GDD/opcode.h>

#include "reduce.h"

#include "stringutil.h"
#include "util.h"

#include "param.h"
#include "solver.h"

/** Reduce register saturation of a given DAG.

    @param[in] ddg = DAG
    @param[in] settings = solver settings
    @param[in] weight_map = weight map giving for each type its weight
    @param[out] added_edges = list of (antidep) edges added to the DAG
    in case of success

    @return -1 if no solution or a positive value if a solution was
    found.
*/
static int SIRA_reduce_internal(SCEDA_Graph *g, SIRA_Param *settings, SIRA_WeightMap *weight_map, SCEDA_List **added_edges) {
  if(!SCEDA_graph_is_acyclic(g)) {
    return -1;
  }

  /** Add source and sink */
  GDD_Opcode *dummy_op = GDD_opcode_create("dummy", 0, 0);
  GDD_NodeLabel *dummy_node = GDD_node_label_create(-1, dummy_op);

  SCEDA_Vertex *v_source = SCEDA_graph_add_vertex(g, dummy_node);
  SCEDA_Vertex *v_sink = SCEDA_graph_add_vertex(g, dummy_node);
  
  SCEDA_List *new_edges = SCEDA_list_create(NULL);
  
  {
    SCEDA_VerticesIterator g_vertices;
    SCEDA_vertices_iterator_init(g, &g_vertices);
    while(SCEDA_vertices_iterator_has_next(&g_vertices)) {
      SCEDA_Vertex *v = SCEDA_vertices_iterator_next(&g_vertices);
      if(v != v_source) {
	if(SCEDA_vertex_in_deg(v) == 0) {
	  GDD_EdgeLabel *e = GDD_edge_label_create(NULL, "serial", 0, 0);
	  SCEDA_Edge *virt_e = SCEDA_graph_add_edge(g, v_source, v, e);
	  safe_call(SCEDA_list_add(new_edges, virt_e));
	}
      }
      if(v != v_sink) {
	if(SCEDA_vertex_out_deg(v) == 0) {
	  GDD_EdgeLabel *e = GDD_edge_label_create(NULL, "serial", GDD_vertex_latency(v), 0);
	  SCEDA_Edge *virt_e = SCEDA_graph_add_edge(g, v, v_sink, e);
	  safe_call(SCEDA_list_add(new_edges, virt_e));
	}
      }
    }
    SCEDA_vertices_iterator_cleanup(&g_vertices);
  }

  /** Add an edge from sink to source */
  {
    GDD_EdgeLabel *e = GDD_edge_label_create(NULL, "serial", 1, 1);
    SCEDA_Edge *virt_e = SCEDA_graph_add_edge(g, v_sink, v_source, e);
    safe_call(SCEDA_list_add(new_edges, virt_e));
  }

  /** Construct SiraLoop */
  SIRA_Loop *loop = SIRA_loop_create(g, -1); 
  int II = SIRA_loop_Lmax(loop); 

  SCEDA_HashMap *result; 

  SCEDA_HashMap *reuse_decisions;
  if(SIRA_solver_single(loop, settings, weight_map, II, &reuse_decisions) != 0) {
    II = -1;
  }
  if(II != -1) {
    result = SCEDA_hashmap_create((SCEDA_delete_fun)free, (SCEDA_delete_fun)SCEDA_list_delete, (SCEDA_match_fun)string_match, (SCEDA_hash_fun)string_hash);
      
    SCEDA_HashMapIterator decisions;
    SCEDA_hashmap_iterator_init(reuse_decisions, &decisions);
    while(SCEDA_hashmap_iterator_has_next(&decisions)) {
      char *type;
      SCEDA_List *reuse_edges = SCEDA_hashmap_iterator_next(&decisions, &type);
      SCEDA_List *antidep_edges = SIRA_reuse_apply_decisions(SIRA_loop_ddg(loop), type, reuse_edges, SIRA_param_get_ual_flag(settings));
      safe_call(SCEDA_hashmap_put(result, safe_strdup(type), antidep_edges, NULL));
    }
    SCEDA_hashmap_iterator_cleanup(&decisions);

    SCEDA_hashmap_delete(reuse_decisions);
  }

  SIRA_loop_delete(loop);

  SCEDA_List *good_edges = NULL;
  if(II != -1) {
    good_edges = SCEDA_list_create(NULL);
    SCEDA_List *bad_edges = SCEDA_list_create(NULL);
    
    SCEDA_HashMapIterator edges;
    SCEDA_hashmap_iterator_init(result, &edges);
    while(SCEDA_hashmap_iterator_has_next(&edges)) {
      char *type;
      SCEDA_List *antidep_edges = SCEDA_hashmap_iterator_next(&edges, &type);
      SCEDA_ListIterator ae;
      SCEDA_list_iterator_init(antidep_edges, &ae);
      while(SCEDA_list_iterator_has_next(&ae)) {
	SCEDA_Edge *e = SCEDA_list_iterator_next(&ae);
	if((SCEDA_edge_source(e) == v_source) || (SCEDA_edge_source(e) == v_sink) ||
	   (SCEDA_edge_target(e) == v_source) || (SCEDA_edge_source(e) == v_sink) ||
	   (GDD_edge_distance(e) != 0)) {
	  safe_call(SCEDA_list_add(bad_edges, e));
	} else {
	  safe_call(SCEDA_list_add(good_edges, e));
	}
      }
      SCEDA_list_iterator_cleanup(&ae);
    }
    SCEDA_hashmap_iterator_cleanup(&edges);

    SCEDA_hashmap_delete(result);

    GDD_delete_edges(g, bad_edges);
    SCEDA_list_delete(bad_edges);
  }

  /** Restore original graph */
  {
    GDD_delete_edges(g, new_edges);
    SCEDA_list_delete(new_edges);

    GDD_NodeLabel *node;
    safe_call(SCEDA_graph_remove_vertex(g, v_source, (void **)&node));
    safe_call(SCEDA_graph_remove_vertex(g, v_sink, (void **)&node));
    GDD_node_label_delete(dummy_node);
    GDD_opcode_delete(dummy_op);
  }

  if(II != -1) {
    if(SCEDA_graph_is_acyclic(g)) {
      *added_edges = good_edges;
    } else {
      II = -2;
      
      GDD_delete_edges(g, good_edges);
      SCEDA_list_delete(good_edges);
    }
  }
  
  return II;
}

/** Reduce register saturation of a given DAG.

    @param[in] ddg = DAG
    @param[in] weight_map = Weight Map giving for each type its weight
    @param[out] added_edges = list of (antidep) edges added to the DAG

    @return -1 if no solution or a positive value if a solution was
    found.
*/
int SIRA_reduce(SCEDA_Graph *ddg, SIRA_WeightMap *weight_map, SCEDA_List **added_edges) {
  SIRA_Param settings;
  SIRA_param_init(&settings);
  SIRA_param_set_ual_flag(&settings, FALSE);
  SIRA_param_set_bounded_flag(&settings, TRUE);
  SIRA_param_set_check_flag(&settings, FALSE);
  int II = SIRA_reduce_internal(ddg, &settings, weight_map, added_edges);
  if(II >= 0) {
    SIRA_param_cleanup(&settings);
    return II;
  }
  SIRA_param_set_ual_flag(&settings, TRUE);
  II = SIRA_reduce_internal(ddg, &settings, weight_map, added_edges);
  SIRA_param_cleanup(&settings);
  if(II < 0) {
    II = -1;
  }
  return II;
}

