/*
   Copyright Universite de Versailles Saint-Quentin en Yvelines 2009
   AUTHORS: Sebastien Briais, Sid Touati

   This file is part of SIRA.
   
   SIRA is free software: you can redistribute it and/or modify it
   under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation, either version 3 of the
   License, or (at your option) any later version.
   
   SIRA is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with SIRA.  If not, see
   <http://www.gnu.org/licenses/>.
*/
#ifndef __SIRA_REDUCE_H
#define __SIRA_REDUCE_H
/** \file reduce.h
    \brief Register saturation reduction heuristic */

#include <SCEDA/graph.h>
#include <SCEDA/list.h>

#include <GDD/arch.h>
#include "weightmap.h"

/** Reduce register saturation of a given DAG.

    @param[in] ddg = DAG
    @param[in] weight_map = Weight Map giving for each type its weight
    @param[out] added_edges = list of (antidep) edges added to the DAG

    @return -1 if error or otherwise a positive value.

    Note that the graph is modified in case of success: edges are
    added to the graph and listed in added_edges.
*/
int SIRA_reduce(SCEDA_Graph *ddg, SIRA_WeightMap *weight_map, SCEDA_List **added_edges);

#endif

/** \page redrs Register saturation reduction heuristic for DAGs

    \section redrs_intro Introduction

    SiraLina can be efficiently used to devise a register saturation
    reduction heuristic for acyclic DDGs.

    The heuristic works as follows. 

    Let \f$G = (V,E,\delta,\lambda)\f$ be a DAG (\f$\lambda = 0\f$).

    \li add a source \f$\top\f$ and a sink \f$\bot\f$ to G and an edge
    \f$e = (\bot,\top)\f$ such that \f$\delta(e) = \lambda(e) = 1\f$

    \li apply SiraLina to the resulting DDG in order to get a set of
    reuse edges

    \li computes the associated DDG but filter out edges \f$e'\f$ such
    that \f$\lambda(e') \neq 0\f$

    \li remove \f$\top\f$, \f$\bot\f$ and \f$e\f$.

    \section redres_api Register saturation reduction heuristic API

    \code
    int SIRA_reduce(SCEDA_Graph *ddg, SIRA_WeightMap *weight_map, SCEDA_List **added_edges);
    \endcode

    Apply the register saturation reduction heuristic to the given
    acyclic DDG.

    \section redres_exa Example

    \include "reduce/main.c"

 */
