/*
   Copyright Universite de Versailles Saint-Quentin en Yvelines 2009
   AUTHORS: Sebastien Briais, Sid Touati

   This file is part of SIRA.
   
   SIRA is free software: you can redistribute it and/or modify it
   under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation, either version 3 of the
   License, or (at your option) any later version.
   
   SIRA is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with SIRA.  If not, see
   <http://www.gnu.org/licenses/>.
*/
#ifndef __SIRA_REUSE_H
#define __SIRA_REUSE_H
/** \file reuse.h
    \brief Reuse edge descriptions */

#include <SCEDA/graph.h>
#include <SCEDA/list.h>
#include <SCEDA/hashmap.h>

/** Type of reuse edges description */
typedef struct {
  /** Source vertex */
  SCEDA_Vertex *source;
  /** Target vertex */
  SCEDA_Vertex *target;
  /** Distance */
  int lambda;
} SIRA_ReuseEdge;

/** Creates a reuse edge description of given characteristics.

    @param[in] source = source vertex 
    @param[in] target = target vertex
    @param[in] lambda = distance 

    @return a reuse edge description */
SIRA_ReuseEdge *SIRA_reuse_edge_create(SCEDA_Vertex *source, SCEDA_Vertex *target, int lambda);

/** Delete a reuse edge description.

    @param[in] e = reuse edge description */
void SIRA_reuse_edge_delete(SIRA_ReuseEdge *e);

/** Computes the cost of a list of reuse edges (sum of all the distances).

    @param[in] reuse_edges = list of reuse edges

    @return the cost */                                                   
int SIRA_reuse_decisions_get_cost(SCEDA_List *reuse_edges);  

/** Apply a set of reuse decisions to the corresponding DDG.
    
    @param[in] g = DDG
    @param[in] type = type
    @param[in] reuse_edges = reuse edges of given type
    @param[in] ual_flag = UAL semantics switch
    
    @return the list of added edges (could be deleted with ddg_delete_edges) */
SCEDA_List *SIRA_reuse_apply_decisions(SCEDA_Graph *g, const char *type, SCEDA_List *reuse_edges, int ual_flag);

#if(0)
/** Type of (antidep) edges description */
typedef struct {
  /** Source vertex */
  SCEDA_Vertex *source;
  /** Target vertex */
  SCEDA_Vertex *target;
  /** Latency */
  int latency;
  /** Distance */
  int distance;
  /** Type */
  char *type;
} SIRA_EdgeDesc;

/** Get the list of antidep edge descriptions that correspond to reuse decisions.

    @param[in] g = DDG
    @param[in] type = type
    @param[in] reuse_edges = reuse edges of given type
    @param[in] ual_flag = UAL semantics switch

    @return the list of (antidep) edge descriptions */
SCEDA_List *SIRA_reuse_decisions_to_edges_desc(SCEDA_Graph *g, const char *type, SCEDA_List *reuse_edges, int ual_flag);
#endif

#endif

/** \page reuse Reuse Edges Map: output of SiraLina solver.

    \section reuse_intro Introduction

    The \ref solver indicates, for each type, which reuse edges should
    be added in the reuse graph. These reuse decisions correspond to
    new edges that are to be added to the original DDG. In the sequel,
    we describe the translation mechanism.

    A reuse edge \f$(u,v)\f$ of type \f$t\f$ and of reuse distance
    \f$\mu_{u,v}^t\f$ implies the addition of the following edges to
    the DDG:

    \li if \f$\mathop{Cons}^t(u) = \emptyset\f$, then an antidependency edge
    of type \f$t\f$ is added from \f$u\f$ to \f$v\f$ with a latency
    \f$\delta = \delta_r^t(u) - \delta_w^t(v)\f$ and a distance
    \f$\lambda = \mu_{u,v}^t\f$.

    \li otherwise, for each consumers \f$c_u \in \mathop{Cons}^t(u)\f$,
    an antidependency edge of type \f$t\f$ is added from \f$c_u\f$ to
    \f$v\f$ with a latency \f$\delta = \delta_r^t(c_u) -
    \delta_w^t(v)\f$ and a distance \f$\lambda = \mu_{u,v}^t -
    \max_{e \in E_R^t \wedge \mathop{source}(e) = u \wedge
    \mathop{target}(e) = c_u} \lambda(e)\f$.

    \section reuse_api Reuse Edges API

    \code
    SIRA_ReuseEdge *SIRA_reuse_edge_create(SCEDA_Vertex *source, SCEDA_Vertex *target, int lambda);
    \endcode

    Create a reuse edge between source vertex and target vertex (in a
    same DDG), with given reuse distance.

    Note that reuse edges do not explicitly have a type since SiraLina
    solution is stored in a map indexed by types.

    \code
    void SIRA_reuse_edge_delete(SIRA_ReuseEdge *e);
    \endcode

    Delete a reuse edge.

    \code
    int SIRA_reuse_decisions_get_cost(SCEDA_List *reuse_edges);
    \endcode

    Compute the sum of the reuse distances of the given list of reuse
    edges.

    \code 
    SCEDA_List *SIRA_reuse_apply_decisions(SCEDA_Graph *g, const char *type, SCEDA_List *reuse_edges, int ual_flag);
    \endcode

    Apply a set of reuse decisions as given by the siralina solver by
    modifying the corresponding DDG. It returns the list of added
    edges (antidependency edges) that corresponds to the list of reuse
    edges of given type. These new edges can/must be removed from g by
    using ddg_delete_edges.

    It follows the translation mechanism described above.
*/

/*
    \code
    SCEDA_List *SIRA_reuse_decisions_to_edges_desc(SCEDA_Graph *g, const char *type, SCEDA_List *reuse_edges, int ual_flag);
    \endcode
    
    Instead of applying reuse decisions to a DDG, it is possible to
    get a description of the antidep edges that will be added to the
    DDG if reuse decisions are applied. This function do this job
    by returning a list of SIRA_EdgeDesc.
 */
